#! /usr/local/bin/perl -w

# 2230, Sat 13 Jan 01 (NZDT)
#
# Nevil Brownlee, CAIDA / U Auckland

# Merges flows with the same attribute values
#
# Command line: 
#   ./at_merge.pl in.fdf out.fdf  flowkind  % toturnaroundtime1
#          from_file to_file  attribs  +%/ distribs required

use strict;
use Time::Local;

local $::ln10 = 2.3025850929940457;

local %::month = (
      "Jan" =>  1, "Feb" =>  2, "Mar" =>  3, "Apr" =>  4,
      "May" =>  5, "Jun" =>  6, "Jul" =>  7, "Aug" =>  8,
      "Sep" =>  9, "Oct" => 10, "Nov" => 11, "Dec" => 12,
   );

my ($c_interval, $samples_per_int, $delta, $first_plot);

my ($from_file, $to_file);

$from_file = shift;  # First item from command line
if (defined($from_file)) {
   open(FROM, $from_file)
   or die "Can't open from_file: $!\n";
   }
else { printf "No from_file specified!\n"; }
 

$to_file = shift;  # Second item from command line
if (defined($to_file)) {
   open(TO, ">>$to_file")
   or die "Can't open to_file: $!\n";
   }
else { printf "No to_file specified!\n"; }

my $ntm_hdr = <FROM>;  # Read the input header record
if (!defined($ntm_hdr)) {
   print "read from $from_file falied!\n";
   exit;
   }
if ($ntm_hdr =~ /-c(\d+) /) {
   $c_interval = $1;
   }
print TO $ntm_hdr;

my $format;
for (;;) {  #  Find the #Format record
   $format = <FROM>;
   print TO $format;
   last if $format =~ /^#Format: /
   }
$format =~ s/^#Format: //;

my ($nf, %format, %dist, $next);
my (@type, @dif);
# type: 0 = scalar, 1 = tcpdata, 2 = distribution
# dif:  1 = dif (to be summed), 0 = fixed

unpack_format_record($format);

my ($c1,$c2);  # c1..c2 columns written to stats file
my ($item, @attrb, @attst, @attvl, @att_str, @attvh, @nvalues, @amult);
my $errs = 0;

for (my $a = 0; $item = shift; ) {  # Grab items from command line
   my ($atr, $test, $val, $str_atr, $valh);

   last if $item eq '+';  # Stop when we find a '+'
   last if $item eq '%';  #    or a '%'
   last if $item eq '/';  #    or a '/'
   $item = lc($item);

   if ($item =~ s/\.\.(\d+)//) { 
      $valh = $1;
      }
   else { $valh = ''; }
   if ($item =~ /^(\w+)([=!\#]+)([\w.]+)/) {
      $atr = $1;  $test = substr($2,0,1);  $val = $3;
      $str_atr = $val =~ /\./ ;
      if ($valh ne '' && $val > $valh) {
         print "Bad attribute range for $atr ($val > $valh)\n";
         ++$errs;
         }
      }
   else { $atr = $item;  $test = '=';  $val = ''; }

   if (defined $format{$atr} ) {
      $attrb[$a] = $atr;  $attst[$a] = $test;
      $attvl[$a] = $val;  $att_str[$a] = $str_atr;
      $attvh[$a] = $valh;
      $nvalues[$a] = $valh ne '' ? $valh-$val+1 : ($val ne '' ? 1 : 0);
      printf " %s %s %s %s (%d values)\n", $atr, $test, $val,
         $valh ne '' ? ".. $valh" : '', $nvalues[$a];
      ++$a;
      }
   else {
      print STDERR "unknown attrib '$atr'\n";
      ++$errs;
      }
   }


my @plot;

if (defined($item)) {
   for (my $p=0; $plot[$p] = shift; ++$p ) { # grab distribs
      $plot[$p] = lc($plot[$p]);
      if (defined $format{$plot[$p]}) { 
         printf " %s", $plot[$p];
         }
      else { 
         print STDERR "Unknown distribution '$plot[$p]' <<<\n";
         ++$errs;
         }
      }
   }
pop @plot;   # Last shift will have returned NULL
print "\n";

if ($errs != 0) { exit; }

print TO "##at_merge  $from_file ";
for (my $j = 0; $j != scalar(@attrb); ++$j) {
   print TO " $attrb[$j]$attst[$j]$attvl[$j]";
   print " key $j: $attrb[$j]$attst[$j]$attvl[$j]";
   if ($attvh[$j] ne '') {
      print TO "..$attvh[$j]";
      print  "..$attvh[$j]";
      }
   print "\n";
   }
print TO "\n";

# %format       = index in records for attribtes (from #Format)
# @attrb,@attvl = array of attrib/value pairs for selecting flows
# $item         = +  Use distrib raw counts (i.e. 'as is')
#                 %  Convert each sample to pecentages
#                 /  Convert to percentages, but ignore oflo bucket
# @plot         = list of attribs to plot

#printf "attrib %s=%s, plot %s=%d\n",
#    $attrb[0], $attvl[0], $plot[0], $format{$plot[0]};

my $nbr = 0;

my $last_c_time = 0;  # Globals for get_time()
my $x;       # x value (time of day) for gnuplot
my $c_time;  # Collection time (seconds from 1970)

my %flows = ();  # Hash for flows in sample

while (<FROM>) { # read  flows file

   chop;    # trim new line
   if (/^#/) { # starts with a #
      my $hrec = $_;
      if (s/^#Time: //) {
         print TO "$hrec\n";  # Copy the # record
 	 } 
      elsif (s/^##plt_stat: //) {
         print TO "$hrec\n";  # Copy the # record
         printf TO "##at_merge.pl  $from_file ";
         for (my $j = 0; $j != scalar(@attrb); ++$j) {
            printf TO " $attrb[$j]$attst[$j]$attvl[$j]";
            }
         print TO "\n";
	 } 
      elsif (s/^#EndData: //) {
         foreach my $key (sort keys %flows) {
            print_flow(\*TO, $flows{$key});
	    }
         print TO "$hrec\n";  # Copy the # record
         %flows = ();
         }
      elsif (s/^#Format: //) {
         $hrec =~ s/^#Format: //;
         unpack_format_record($hrec);
         for (my $p = 0; $p != scalar(@plot); ++$p ) { # Grab distribs
            if (!defined $format{$plot[$p]}) { 
               print STDERRrec
        "Format change, last time $x: nknown distribution '$plot[$p]' <<<\n";
               ++$errs;
               }
            }
         }
      else {
         print TO "$hrec\n";  # Copy the # record
         }
      next;
      }

   my $vref = get_flow();

   my $key = match_flow($vref);  # '' if we don't want the flow
   if ($key ne '') {
      my $aref = $flows{$key};
      if (!defined($aref)) {
         $flows{$key} = $vref;
         }
      else {  # Already seen this key
         accumulate_flow($aref, $vref);
         }
      }
#   ++$nbr;
#   if ($nbr > 90) { exit; }
   }

# That's it
#--------------------------------------------------------------------------

sub unpack_format_record {
   my ($fmt) = @_;

   $nf = 0;  %format = ();  %dist = (); @type = ();  @dif  = ();

   for ($nf = 0; $fmt =~ s/([\w\[\(]\w*[\]|\)]?)//; ++$nf) {  # Split up fmt
      my $next = $1;
      $type[$nf] = 0;
      if ($next =~ s/^\((.+)\)$/$1/ ) {
         $dist{$next} = $nf;
         $type[$nf] = 2;  # Distribution
         }
      elsif ($next =~ s/^\[(.+)\]$/$1/ ) {
         $type[$nf] = 1;  # TCPdata
         }
      $format{$next} = $nf;
      $dif[$nf] = 0;
      if ($next =~ /^d_/) {
         $dif[$nf] = 1;  # To be summed
         }
      }
   }


sub match_flow{  # Returns flow key, '' if match fails
   my ($vref) = @_;

   my ($av, $match, $fail, $key);
   $fail = 0;  $key = '';
   for (my $q = 0; $q != scalar(@attrb); ++$q) {
      $av = $vref->[$format{$attrb[$q]}];
      $key = $key . "$av ";
      if ($attvl[$q] ne '') {
         if ($attvh[$q] eq '') {
            $match = $att_str[$q] ? $av eq $attvl[$q] : $av == $attvl[$q];
            }
         else {
            $match = ($av >= $attvl[$q] && $av <= $attvh[$q]);
            }
         if ($match) {
            last if $fail = $attst[$q] eq '!';
            }
         else {
            last if $fail = $attst[$q] eq '=';
            }
         } 
      }
   return $fail ? '' : $key;
   }

sub get_flow{  # Take $_, return ref to array of values

   my @value = ();
   for (my $i = 0;  s/([0123456789ABCDEF.-]+|\(|\[)//;  $i++) {
         # remove leading non-numerics followed by (, [ or a value
         # NOTE: .- appear in addresses, : appears in V6 addresses
      if ($1 eq '(') {  #  have distrib
	 s/\W*([^)]+)\)//;  # remove everything upto next ) save it in $1
	 my @v = split(/ +/, $1);   # put the distribution values in @v
	 my $v = {};                # $v points to a hash
	 if (($v->{Trans} = $v[0]) != 0 ) {  # construct structure
	    $v->{Scale} = $v[1];
	    $v->{Lower} = $v[2];
	    $v->{Upper} = $v[3];
	    $v->{Buckets} = $v[4];
	    $v->{P1} = $v[5];
	    $v->{P2} = $v[6];
	    $v->{P3} = $v[7];
	    $v->{Dist} = [splice(@v,8)]; # copies from element 8 to end
#            if ($v->{P1} != 0) {
#               $samples_per_int = $c_interval/$v->{P1};
#               $v->{Dist}->[0] %= $samples_per_int;
#                  # Ignore empty collection intervals
#               }
            $value[$i] = $v;
	    }
	 } 
      elsif ($1 eq '[') {  # TCPdata
	 s/([^\]]+)\]//;  # Remove everything upto next ], save it in $1 
         my @v = split(/ +/, $1);
	 $value[$i] = \@v;
         }
      else {
	 $value[$i] = $1;
	 }
      }
   return \@value;  # Return ref to array
   }
 

sub accumulate_flow{  # aref += vref
   my ($aref, $vref) = @_;

   for (my $j = 0; $j != $nf; ++$j) {
      if ($type[$j] == 0) {  # Scalar
         if ($dif[$j] == 1) {
            $aref->[$j] += $vref->[$j];
            }
         }
      elsif ($type[$j] == 2) {  # Distribution
         my $ar = $aref->[$j];
         my $vr = $vref->[$j];
         if ($dif[$j] == 1) {
            if (!defined($ar->{Trans})) {
               $aref->[$j] = $vref->[$j];
	       }
            elsif (defined($vr->{Trans})) {
               my @abucket = @{ $ar->{Dist} };
               my @vbucket = @{ $vr->{Dist} };
               for (my $k = 0; $k != scalar(@abucket); ++$k) {
                  $abucket[$k] += $vbucket[$k];
                  }
               $ar->{Dist} = \@abucket;
               }
            }
         }
      else {  # TCPdata
         my @a = $aref->[$j];
         my @v = $vref->[$j];
         if ($dif[$j] == 1) {
            if (!defined(@a)) {
               $aref->[$j] = $vref->[$j];
	       }
            else {
               $a[0] += $v[0];  # Nbr of streams
               $a[1] = $v[1] > $a[1] ? 
                  $v[1] : $a[1];  # Max active atreams
               for (my $k = 2; $k != scalar(@a); ++$k) {
                  $a[$k] += $v[$k];
                  }
               }
            }
         }
      }
   }


sub print_flow{  # Write output record
   my ($FILE, $vref) = @_;

   for (my $j = 0; $j != $nf; ++$j) {
      if ($type[$j] == 0) {  # Scalar
         print $FILE "$vref->[$j] ";
         }
      elsif ($type[$j] == 2) {  # Distribution
         my $hr = $vref->[$j];
         if (!defined($hr->{Trans})) {
            print $FILE "(0) ";  # Empty distribution
	    }
	 else {
            print $FILE 
               "($hr->{Trans} $hr->{Scale} $hr->{Lower} $hr->{Upper}  ";
            print $FILE 
               "$hr->{Buckets} $hr->{P1} $hr->{P2} $hr->{P3}  ";
            my @bucket = @{ $hr->{Dist} };
            print $FILE "@bucket) ";
            }
         }
      else {  # TCPdata
         my @data = @{ $vref->[$j] };
         print $FILE "[@data] ";
         }
      }
   print $FILE "\n";
   }


sub get_time{  # Take #Time from $_, return $x, $delta, $last_ctime
   if (/(^\d+):(\d+):(\d+) +\w+ +(\d+) +(\w+) +(\d+) +/) {
      #    HH    MM   SS    day    dd    mm     yyyy
#      $c_time = timegm($3,$2,$1, $4,$::month{$5}-1,$6-1900);
      $c_time = timelocal($3,$2,$1, $4,$::month{$5}-1,$6-1900);
#      print $_, "  ", $c_time, " ", $last_c_time, "\n";
      if ($last_c_time != 0) {
         my $interval = $c_time - $last_c_time;
         $delta = $interval - $c_interval;
         if ($delta < -2 || $delta > 2) {
            print STDERR $_, "\n";
            print STDERR "   interval = $interval <<<<<<<<<\n";
            }
         }
      $last_c_time = $c_time;
      $x = sprintf("%02d%02d.%02d%02d", $::month{$5},$4, $1,$2);
      }               # mmdd.HHMM
#      $x = sprintf("%04d%02d%02d:%02d%02d", $6,$::month{$5},$4, $1,$2);
#      }               # YYYYmmdd:HHMM
   }
